/* crlcache.c - LDAP access
 *      Copyright (C) 2002 Klarlvdalens Datakonsult AB
 *
 * This file is part of DirMngr.
 *
 * DirMngr is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DirMngr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <assert.h>
#include <dirent.h>
#include <netinet/in.h>

#include "dirmngr.h"
#include "crlcache.h"
#include "cert.h"
#include "crlfetch.h"
#include "misc.h"

#define fail_if_err(a) do { if(a) {                                       \
                              fprintf (stderr, "%s:%d: KSBA error: %s\n", \
                              __FILE__, __LINE__, ksba_strerror(a));   \
                              exit (1); }                              \
                           } while(0)


#define fail_if_err2(f, a) do { if(a) {\
            fprintf (stderr, "%s:%d: KSBA error on file `%s': %s\n", \
                       __FILE__, __LINE__, (f), ksba_strerror(a));   \
                            exit (1); }                              \
                           } while(0)

#define fail(s)  do { fprintf (stderr, "%s:%d: %s\n", __FILE__,__LINE__, (s));\
                      exit (1); } while(0)

static int mydbopen( const char* file, DB** dbpp )
{
  return db_open( file, DB_HASH, DB_CREATE, S_IRUSR|S_IWUSR, NULL, NULL, dbpp );
}

static int mydbclose( DB* dbp )
{
  return dbp->close( dbp, 0 );
}

static int mydbget( DB* dbp, DBT *key, DBT *data )
{
  return dbp->get( dbp, NULL, key, data, 0 );
}

static int mydbput( DB* dbp, DBT *key, DBT *data )
{
  return dbp->put( dbp, NULL, key, data, 0 );
}

static void* make_db_val( time_t t, unsigned char reason )
{
  char* val = xmalloc( sizeof( u32 ) + sizeof( unsigned char ) );
  DIRMNGR_START_FNC;
  ((u32)*val) = (u32)htons( t );
  ((unsigned char)*(val+sizeof(u32))) = reason;
  return val;
}

void parse_db_val( const void* val, time_t* t, unsigned char* reason )
{
  const char* _val = val;
  DIRMNGR_START_FNC;
  *t = ntohs((u32)*_val);
  *reason = ((unsigned )*(_val+sizeof(u32)));  
}

#if 0
static void
print_sexp (KsbaConstSexp p)
{
  int level = 0;

  if (!p)
    fputs ("[none]", stderr);
  else
    {
      for (;;)
        {
          if (*p == '(')
            {
              fputc (*p, stderr);
              p++;
              level++;
            }
          else if (*p == ')')
            {
              fputc (*p, stderr);
              if (--level <= 0 )
                return;
            }
          else if (!digitp (p))
            {
              fputs ("[invalid s-exp]", stderr);
              return;
            }
          else
            {
              KsbaConstSexp endp;
              unsigned long n;

              n = strtoul (p, (char**)&endp, 10);
              p = endp;
              if (*p != ':')
                {
                  fputs ("[invalid s-exp]", stderr);
                  return;
                }
              fputc('#', stderr);
              for (p++; n; n--, p++)
                fprintf (stderr, "%02X", *p);
              fputc('#', stderr);
            }
        }
    }
}

static void
print_dn (char *p)
{
  if (!p)
    fputs ("error", stderr);
  else
    log_debug( "`%s'", p);
}

#endif

static void
print_time (time_t t, FILE* fp)
{

  if (!t)
    fputs ("none", fp);
  else if ( t == (time_t)(-1) )
    fputs ("error", fp);
  else
    {
      struct tm *tp;

      tp = gmtime (&t);
      fprintf( fp, "%04d-%02d-%02d %02d:%02d:%02d",
              1900+tp->tm_year, tp->tm_mon+1, tp->tm_mday,
              tp->tm_hour, tp->tm_min, tp->tm_sec);
      assert (!tp->tm_isdst);
    }
}

static CrlCache* crl_cache_new(void)
{
  CrlCache* c = (CrlCache*)xmalloc( sizeof( CrlCache ) );
  FILE* contents;
  char* fname = make_filename( dirmngr_opt.homedir, DBCONTENTSFILE, NULL );
  char buf[1024];
  int rc;

  DIRMNGR_START_FNC;
  contents = fopen( fname, "r" );
  if( contents == NULL ) {
    log_error( "Error opening DB contents file %s", fname );
  }
  gcry_free( fname );
  c->entries = NULL;

  {
    /* Make sure cache directory exists */
    DIR* dir;
    fname = make_filename( dirmngr_opt.homedir, NULL );
    dir = opendir( fname );
    if( dir == NULL ) {
      log_debug( "Creating directory %s\n", fname );
      if( mkdir( fname, S_IRUSR|S_IWUSR|S_IXUSR ) ) {
	perror( "Error creating directory");
      }
    } else {
      closedir( dir );
    }
    gcry_free( fname );
    fname = make_filename( dirmngr_opt.homedir, DBDIR, NULL );
    dir = opendir( fname );
    if( dir == NULL ) {
      log_debug( "Creating directory %s\n", fname );
      if( mkdir( fname, S_IRUSR|S_IWUSR|S_IXUSR ) ) {
	perror( "Error creating directory");
      }
    } else {
      closedir( dir );
    }
    gcry_free( fname );
  }
  /* Read file version number */
  if( contents && fgets( buf, sizeof( buf ), contents ) ) {
    int len = strlen( buf );
    if( buf[ len-1 ] == '\n' ) buf[ len-1 ] = '\0';
    if( strcmp( buf, DBCONTENTSVERSION ) ) {
      /* Not same version, dont read file */
      fclose( contents );
      contents = NULL;
      /* Clean up cache dir */
      {
	char* dname = make_filename( dirmngr_opt.homedir, DBDIR, NULL );
	DIR* dir = opendir( dname );
	if( dir ) {
	  struct dirent* de;
	  while( ( de = readdir( dir ) ) != NULL ) {
	    if( strcmp( de->d_name, "." ) && strcmp( de->d_name, ".." ) ) {
	      char* dbfname = make_filename( dname, de->d_name, NULL );
	      log_debug( "Deleting %s\n", dbfname );
	      unlink( dbfname ); /* we just ignore errors */	    
	      xfree( dbfname );
	    }
	  }	  
	}
	xfree( dname );
      }
    }
  }
  while( contents && !feof( contents ) ) {
    /*int ok = 1;*/
    CrlCacheEntry* entry = (CrlCacheEntry*)xmalloc( sizeof( CrlCacheEntry ) );
    entry->next = NULL;
    entry->issuer = NULL;
    entry->issuer_hash = NULL;
    /* Read URL */
    if( fgets( buf, sizeof( buf ), contents ) ) {
      int len = strlen( buf );
      if( buf[ len-1 ] == '\n' ) buf[ len-1 ] = '\0';
      entry->url = xstrdup( buf );
    } else {
      xfree( entry );
      break;
    }
    /* Read DN */
    if( fgets( buf, sizeof( buf ), contents ) ) {
      int len = strlen( buf );
      if( buf[ len-1 ] == '\n' ) buf[ len-1 ] = '\0';
      entry->issuer = xstrdup( buf );
    } else {
      xfree( entry );
      break;
    }
    /* Read DN hash */
    if( fgets( buf, sizeof( buf ), contents ) ) {
      int len = strlen( buf );
      if( buf[ len-1 ] == '\n' ) buf[ len-1 ] = '\0';
      entry->issuer_hash = xstrdup( buf );
    } else {
      xfree( entry );
      break;
    }
    /* Read this update time */
    if( fgets( buf, sizeof( buf ), contents ) ) {
      entry->this_update = iso_string2time( buf );
    }    
    /* Read next update time */
    if( fgets( buf, sizeof( buf ), contents ) ) {
      entry->next_update = iso_string2time( buf );
    }
    fname = make_filename( dirmngr_opt.homedir, DBDIR, entry->issuer_hash, NULL );
    log_debug( "Opening DB %s", fname );
    /*entry->dbf = db_open( fname, O_CREAT|O_RDWR, S_IRUSR|S_IWUSR, DB_HASH, NULL );*/
    rc = mydbopen( fname, &(entry->dbf) );
    if( rc ) {
      log_error( "Error opening DB (rc=%d): %s", rc, strerror(errno) );
      exit( -1 );
    }
    entry->next = c->entries;
    c->entries = entry;
    gcry_free( fname );
  }
  if( contents != NULL ) fclose( contents );
  return c;
}

static CrlCacheEntry* find_entry( CrlCacheEntry* first, const char* hash )
{
  DIRMNGR_START_FNC;
  while( first != 0 && strcmp( hash, first->issuer_hash ) != 0 ) first = first->next;
  if( first ) log_debug("find_entry found entry %p (issuer=%s)\n", first, first->issuer_hash );
  else log_debug("find_entry found no entry\n");
  return first;    
}

static void save_contents( CrlCacheEntry* first )
{
  CrlCacheEntry* entry = first;
  char* fname = make_filename( dirmngr_opt.homedir, DBCONTENTSFILE, NULL );
  char* fnametmp = make_filename( dirmngr_opt.homedir, DBCONTENTSFILE ".tmp", NULL );
  FILE* contents = fopen( fnametmp, "w" );
  DIRMNGR_START_FNC;
  if( contents == NULL ) {
    log_error( "Error writing contents file: %s", strerror(errno) );
    gcry_free( fname );
    gcry_free( fnametmp );
    dirmgr_exit( 2 );
  }
  fprintf( contents, "%s\n", DBCONTENTSVERSION );
  while( entry ) {
    /*fprintf( stderr, "Writing entry for %s\n", entry->issuer_hash );*/
    fprintf( contents, "%s\n", entry->url );
    fprintf( contents, "%s\n", entry->issuer );
    fprintf( contents, "%s\n", entry->issuer_hash );
    fprintf( contents, "%s\n", iso_time2string( entry->this_update ) );
    fprintf( contents, "%s\n", iso_time2string( entry->next_update ) );
    entry = entry->next;
  }
  fclose( contents );
  if( rename( fnametmp, fname ) ) {
    log_error( "Error moving contents file in place: %s", strerror( errno ) );
  }
  gcry_free( fname );
  gcry_free( fnametmp );
}

CrlCache* crl_cache_init(void)
{
  return crl_cache_new();
}

void crl_cache_deinit( CrlCache* c )
{
  CrlCacheEntry* entry = c->entries;  
  save_contents( c->entries );
  DIRMNGR_START_FNC;
  while( entry ) {
    CrlCacheEntry* tmp;
    if( entry->issuer_hash != NULL ) {
      xfree( entry->issuer_hash );
      entry->issuer_hash = NULL;
    }
    /*entry->dbf->close( entry->dbf, 0 );*/
    mydbclose( entry->dbf );
    tmp = entry->next;
    xfree( entry );
    entry = tmp;    
  }
  xfree( c );
}

static int
do_encode_md (GCRY_MD_HD md, int algo,  unsigned int nbits,
              GCRY_MPI *r_val)
{
  int nframe = (nbits+7) / 8;
  byte *frame;
  int i, n;
  byte asn[100];
  size_t asnlen;
  size_t len;

  DIRMNGR_START_FNC;

  asnlen = DIM(asn);
  if (gcry_md_algo_info (algo, GCRYCTL_GET_ASNOID, asn, &asnlen))
    {
      log_error ("No object identifier for algo %d\n", algo);
      return CRL_SIG_ERROR;
    }

  len = gcry_md_get_algo_dlen (algo);

  if ( len + asnlen + 4  > nframe )
    {
      log_error ("can't encode a %d bit MD into a %d bits frame\n",
                 (int)(len*8), (int)nbits);
      return CRL_SIG_ERROR;
    }

  /* We encode the MD in this way:
   *
   *       0  A PAD(n bytes)   0  ASN(asnlen bytes)  MD(len bytes)
   *
   * PAD consists of FF bytes.
   */
  frame = xtrymalloc (nframe);
  if (!frame)
    return CRL_SIG_ERROR;
  n = 0;
  frame[n++] = 0;
  frame[n++] = 1; /* block type */
  i = nframe - len - asnlen -3 ;
  assert ( i > 1 );
  memset ( frame+n, 0xff, i ); n += i;
  frame[n++] = 0;
  memcpy ( frame+n, asn, asnlen ); n += asnlen;
  memcpy ( frame+n, gcry_md_read(md, algo), len ); n += len;
  assert ( n == nframe );
  if (0)
    {
      int j;
      log_debug ("encoded hash:");
      for (j=0; j < nframe; j++)
        log_printf (" %02X", frame[j]);
      log_printf ("\n");
    }

  gcry_mpi_scan (r_val, GCRYMPI_FMT_USG, frame, &nframe);
  xfree (frame);
  return CRL_SIG_OK;
}

enum CrlCacheReturn crl_cache_isvalid( CrlCache* cptr, const char* issuer_hash, const char* cert_id )
{
  DBT key;
  DBT result;
  int rc;
  CrlCacheEntry* entry = find_entry( cptr->entries, issuer_hash );

  DIRMNGR_START_FNC;
  if( entry == NULL ) {
    /* No CRL from this issuer!! */
    log_info ("no CRL available for issuer `%s'\n", issuer_hash );
    return CRL_CACHE_DONTKNOW;
  }
  if( entry->next_update < time(NULL) ) {
    /* CRL too old, we need to update */
    log_info ("CRL for issuer `%s' too old, we need to update\n", issuer_hash);
    return CRL_CACHE_DONTKNOW;
  }

  memset( &key, 0, sizeof(key) );
  memset( &result, 0, sizeof(result) );

  key.data = (char*)cert_id;
  key.size = strlen( cert_id );
  /*if(  entry->dbf->get( entry->dbf, &key, &result, 0 ) == 1 ) {*/
  rc = mydbget( entry->dbf, &key, &result );
  switch( rc ) {
    case DB_NOTFOUND: {
    log_info( "%s is valid, it is not listed in the CRL\n", cert_id );
    return CRL_CACHE_VALID;
  }
  case 0: {
    time_t when;
    char reason;
    parse_db_val( result.data, &when, &reason );
    {
      struct tm* when_tm = localtime( &when );
      log_info ("cache entry for CRL of `%s' invalid (%lu -- %s,%d)\n",
		issuer_hash, (unsigned long)when, asctime( when_tm ), reason);
    }
    return CRL_CACHE_INVALID;
  }
  default: {
    log_error( "DB->get() error: %s\n", strerror(rc) );
    return CRL_CACHE_DONTKNOW;
  }
  }
}

static int start_sig_check( KsbaCRL crl, GCRY_MD_HD *md, int* algo ) 
{
  const char* algoid = ksba_crl_get_digest_algo (crl);
  DIRMNGR_START_FNC;
  log_debug ( "digest algo: %s\n", algoid? algoid : "[none]");
  
  *algo = gcry_md_map_name ( algoid);
  if (!*algo) {
    log_error ("unknown hash algorithm `%s'\n", algoid? algoid:"?");
    return -1;
  }
  *md = gcry_md_open( *algo, 0 );
  if (!*md) {
    log_error ("md_open failed: %s\n", gcry_strerror (-1));
    return -1;
  }
  
  ksba_crl_set_hash_function(crl, HASH_FNC, *md);  
  return 0;
}

extern KsbaCert get_issuer_cert_local( ASSUAN_CONTEXT ctx, const char* ser );

static KsbaCert get_issuer_cert( ASSUAN_CONTEXT ctx, const char* issuer )
{
  KsbaCert issuer_cert = 0;
  char* cvalue;
  size_t cvaluelen;
  int rc;
  DIRMNGR_START_FNC;
  /* first, ask gpgsm */
  issuer_cert = get_issuer_cert_local( ctx, issuer );
  if( issuer_cert ) {
    return issuer_cert;
  }
  log_debug("No result from inquire\n");
  /* Get issuer cert from server */
  rc = ca_cert_fetch( issuer, &cvalue, &cvaluelen );
  if( rc ) {
    log_error ("error fetching certificate for issuer: rc=%d\n", rc );
    return NULL;
  }
  issuer_cert = ksba_cert_new();
  rc = ksba_cert_init_from_mem( issuer_cert, cvalue, cvaluelen );
  if( rc ) {
    log_error ("invalid issuer certificate: %s\n", ksba_strerror (rc) );
    ksba_cert_release( issuer_cert );
    return NULL;
  }
  else
    log_info ("got issuer certificate\n");

  xfree( cvalue );
  return issuer_cert;
}

static CrlCacheEntry* 
db_entry_new( char* url, char* issuer, char* issuer_hash, 
	      time_t thisupdate, time_t nextupdate )
{
  char* fname;
  CrlCacheEntry* entry;
  int rc;

  DIRMNGR_START_FNC;

  entry = xmalloc( sizeof( CrlCacheEntry ) );
  entry->url = url;
  entry->issuer = issuer;
  entry->issuer_hash = issuer_hash;
  fname = make_filename( dirmngr_opt.homedir, DBDIR, entry->issuer_hash, NULL );
  log_debug( "db_entry_new() fname=%s\n", fname );
  entry->this_update = thisupdate;
  entry->next_update = nextupdate;
  rc = mydbopen( fname, &(entry->dbf) );
  gcry_free( fname );  
  if( rc ) {
    log_error( "Error opening DB (rc=%d): %s\n", rc, strerror( errno ));
    xfree( entry );
    entry = NULL;
    return NULL;
  }
  entry->next = NULL;
  return entry;
}

struct CrlEntry_s {
  char* serial_hex;
  time_t when;
  char reason;
  struct CrlEntry_s* next;
};

typedef struct CrlEntry_s CrlEntry;

static int crl_parse_insert( CrlCache* cptr, const char* url, KsbaCRL crl, ASSUAN_CONTEXT ctx /*for inquire issuer*/  )
{  
  KsbaStopReason stopreason;
  KsbaError err;
  CrlCacheEntry* entry = NULL;
  int new_entry = 0;
  KsbaCert issuer_cert = NULL;
  GCRY_MD_HD md = NULL;
  int rc, algo = 0;
  GCRY_MPI frame;
  KsbaSexp p;
  size_t n;
  GCRY_SEXP s_sig, s_hash, s_pkey;
  CrlEntry* crl_entries = NULL;
  int sig_ok = 0;
  DIRMNGR_START_FNC;

  do {
    err = ksba_crl_parse(crl, &stopreason);
    if( err ) {
      log_error( "ksba_crl_parse error: %s\n", ksba_strerror( err ) );
      /* TODO: clean up! */
      return err;
    }
    switch (stopreason) {
    case KSBA_SR_BEGIN_ITEMS:
      {
	char* issuer = NULL;
	char* issuer_hash = NULL;
	time_t thisupdate; 
	time_t nextupdate; 
	
	if( start_sig_check( crl, &md, &algo ) ) {
	  log_debug( "start_sig_check error %d\n", err );	  
	}
	
	err = ksba_crl_get_issuer (crl, &issuer);
	if( err ) {
	  log_debug( "ksba_crl_get_issuer error %d\n", err );
	}

	issuer_cert = get_issuer_cert( ctx, issuer );
	if( issuer_cert == NULL ) {
	  /* ??? */
	  log_debug( "Could not cert CRL issuer cert!!!\n");
	  DIRMNGR_END_FNC;
	  return CRL_SIG_ERROR;
	}

	issuer_hash = hashify_data( issuer, strlen( issuer ));
	log_debug( "issuer = %s\n", issuer );	
        new_entry = 0;
	entry = find_entry( cptr->entries, issuer_hash );
	log_debug( "entry = %p\n",  entry);	
	err = ksba_crl_get_update_times (crl, &thisupdate, &nextupdate);
	if( err ) {
	  log_debug( "Error in ksba_crl_get_update_times()  %d\n",  err);	  
	}
	log_debug( "update_times are %lu and %lu\n", 
                   (unsigned long)thisupdate, (unsigned long)nextupdate);   
	if( entry == NULL ) {
	  entry = db_entry_new( xstrdup(url), xstrdup(issuer), issuer_hash, thisupdate, nextupdate );
          new_entry = 1;
	} else {
	  log_debug( "Found entry %s\n", entry->issuer_hash);
	  xfree( issuer_hash );
	  issuer_hash = NULL;
	  entry->this_update = thisupdate;
	  entry->next_update = nextupdate;
	}
	log_debug( "Done creating/updating entry\n");	

	ksba_free(issuer);
	log_debug( "freed issuer\n");

	if (err != KSBA_Invalid_Time)
	  fail_if_err2 ("Error", err);
      }
      break;
      
    case KSBA_SR_GOT_ITEM:
      {
	KsbaSexp serial;
	time_t rdate;
	KsbaCRLReason reason;
	CrlEntry* ce = xmalloc( sizeof( CrlEntry ) );
	CrlEntry* tmp_ce;

	err = ksba_crl_get_item (crl, &serial, &rdate, &reason);
	fail_if_err2 ("Error", err);
	
	/*
	  fprintf (stderr, "CRL entry: s=");
	print_sexp (serial);
	fprintf (stderr, " (%s), t=", serial);
	print_time (rdate);
	fprintf (stderr, ", r=%d\n", reason);
	*/
	ce->serial_hex = serial_hex( serial );
	ce->when = rdate;
	ce->reason = reason;
	/*fprintf (stderr, "New CrlEntry (%s) ", ce->serial_hex);*/
	
	tmp_ce = crl_entries;
	crl_entries = ce;
	ce->next = tmp_ce;

	ksba_free( serial );
      }
      break;
      
    case KSBA_SR_END_ITEMS:
      break;
      
    case KSBA_SR_READY:
      {
	gcry_md_final (md);

	p = ksba_crl_get_sig_val (crl);
	n = gcry_sexp_canon_len (p, 0, NULL, NULL);
	if (!n) {
	    log_debug("libksba did not return a proper S-Exp\n");
	    DIRMNGR_END_FNC;
	    return CRL_SIG_ERROR;
	}
	rc = gcry_sexp_sscan ( &s_sig, NULL, p, n);
	if (rc) {
	  log_debug("gcry_sexp_scan failed: %s\n", gcry_strerror (rc));
	  DIRMNGR_END_FNC;
	  return CRL_SIG_ERROR;
	}
	
	p = ksba_cert_get_public_key (issuer_cert);
	n = gcry_sexp_canon_len (p, 0, NULL, NULL);
	if (!n) {
	  log_debug("libksba did not return a proper S-Exp\n");
	  DIRMNGR_END_FNC;
	  return CRL_SIG_ERROR;
	}
	rc = gcry_sexp_sscan ( &s_pkey, NULL, p, n);
	if (rc) {
	  log_debug("gcry_sexp_scan failed: %s\n", gcry_strerror (rc));
	  DIRMNGR_END_FNC;
	  return CRL_SIG_ERROR;
	}
	rc = do_encode_md (md, algo, gcry_pk_get_nbits (s_pkey), &frame);
	if (rc) {
	  /* fixme: clean up some things */
	  DIRMNGR_END_FNC;
	  return rc;
	}
	/* put hash into the S-Exp s_hash */
	if ( gcry_sexp_build (&s_hash, NULL, "%m", frame) )
	  BUG ();
	
	rc = gcry_pk_verify (s_sig, s_hash, s_pkey);
	if (/*DBG_CRYPTO*/0)
	  log_debug("gcry_pk_verify: %s\n", gcry_strerror (rc));
	if( issuer_cert ) ksba_cert_release( issuer_cert );
	/*if( rc ) return CRL_SIG_NOT_OK;
	  else return CRL_SIG_OK;*/
	if( rc ) sig_ok = 0;
	else sig_ok = 1;
      }
      break;
      
    default:
      log_debug( "unknown stop reason\n");
      break;
    }
  } while (stopreason != KSBA_SR_READY);

  if( sig_ok ) {
    /* Signature OK */
    if( entry != NULL ) {
      CrlEntry* e= crl_entries;
      log_debug( "Everything OK, inserting CRL %s\n", entry->issuer_hash);
      while( e != 0 ) {
	CrlEntry* tmp_e = e;
	DBT key, val;   
	int retcode;
	
	log_debug( "Creating db-entry %s -> (%lu, %d)\n", e->serial_hex, e->when, e->reason);

	memset( &key, 0, sizeof(key) );
	memset( &val, 0, sizeof(val) );

	key.data = e->serial_hex;
	key.size = strlen( key.data );
	val.data = make_db_val( e->when, e->reason );
	val.size = sizeof( u32 ) + sizeof( char );
	
	/*retcode = entry->dbf->put( entry->dbf, &key, &val, 0 );*/
	retcode = mydbput( entry->dbf, &key, &val );
	
	xfree( key.data );
	xfree( val.data );      
	tmp_e = e->next;
	xfree( e );
	e = tmp_e;
      }
      if( !find_entry( cptr->entries, entry->issuer_hash ) ) {
	entry->next = cptr->entries;
	cptr->entries = entry;
      }
      save_contents( cptr->entries );
      DIRMNGR_END_FNC;
      if( entry->next_update < time(NULL) ) {
	log_debug( "CRL too old, it expired on %s", ctime( &entry->next_update ));
	return CRL_TOO_OLD;
      }
      return CRL_SIG_OK;
    } else {
      /* ??? */
      log_debug( "Everything not OK: entry missing\n");
    }
  } else {
    /* Signature NOT OK(!) */
    CrlEntry* e= crl_entries;
    log_debug( "Everything not OK: Signature invalid\n");
    while( e != 0 ) {
      CrlEntry* tmp_e = e;
      xfree( e->serial_hex );
      tmp_e = e;
      e = e->next;
      xfree( tmp_e );
    }
    
    if (new_entry) {
      /*entry->dbf->close( entry->dbf );*/
      mydbclose( entry->dbf );
      xfree( entry->url );
      xfree( entry->issuer );
      xfree( entry->issuer_hash );
      xfree( entry );
    }
  }
  /* If we get here, something must be missing
  * so we have an error */
  DIRMNGR_END_FNC;
  return CRL_SIG_NOT_OK; 
}


int crl_cache_insert( CrlCache* cptr, const char* url, const char* value, size_t valuelen, ASSUAN_CONTEXT ctx )
{
  KsbaCRL crl = ksba_crl_new();
  KsbaReader reader = ksba_reader_new();
  int rc;
  DIRMNGR_START_FNC;

  rc = ksba_reader_set_mem( reader, value, valuelen );
  if( rc ) {
    log_debug( "ksba_reader_set_mem %d\n", rc);
    ksba_reader_release( reader );
    ksba_crl_release( crl );
    return -1;
  }
  rc = ksba_crl_set_reader( crl, reader );
  if( rc ) {
    log_debug(  "ksba_reader_set_reader %d\n", rc);
    ksba_reader_release( reader );
    ksba_crl_release( crl );
    return -1;
  }

  rc = crl_parse_insert( cptr, url, crl, ctx );
  ksba_crl_release( crl );
  ksba_reader_release( reader );
  if( rc != CRL_SIG_OK ) {
    const char* c = "";    
    switch( rc ) {
    case CRL_SIG_ERROR:
      c = "CRL_SIG_ERROR"; break;
    case CRL_SIG_NOT_OK:
      c = "CRL_SIG_NOT_OK"; break;
    case CRL_TOO_OLD:
      c = "CRL_TOO_OLD"; break;
    }
    log_debug( "crl_parse_insert %s\n", c);
    DIRMNGR_END_FNC;
    return rc;
  }

  save_contents( cptr->entries );
  DIRMNGR_END_FNC;
  return 0;
}

static int 
list_one_crl_entry( CrlCacheEntry* e, FILE* fp )
{
  DBT key;
  DBT val;
  DBC* cursorp;
  int rc;

  fputs( "--------------------------------------------------------\n", fp );
  fprintf( fp, "Info about CRL from %s\n", e->url );
  fprintf( fp, " Issuer:\t%s\n", e->issuer );
  fprintf( fp, " Issuer Hash:\t%s\n", e->issuer_hash );
  fputs( " Valid from ", fp ); print_time( e->this_update, fp );
  fputs( " to ", fp ); print_time( e->next_update, fp ); fputs( "\n\n", fp );

#ifdef DB2_CURSOR_4_PARAMS
  rc = e->dbf->cursor( e->dbf, NULL, &cursorp, 0 );
#else
  rc = e->dbf->cursor( e->dbf, NULL, &cursorp );
#endif
  if( rc ) return rc;
  memset( &key, 0, sizeof(key) );
  memset( &val, 0, sizeof(val) );
  rc = cursorp->c_get( cursorp, &key, &val, DB_FIRST );
  if( rc ) return rc;  
  do {
    time_t t;
    char reason;
    parse_db_val( val.data, &t, &reason);

    ((char*)key.data)[key.size] = 0;
    fprintf( fp, "  %s:\t reasons( ", (char*)key.data );
    if( reason & KSBA_CRLREASON_UNSPECIFIED ) fputs( "unspecified ", fp );
    if( reason & KSBA_CRLREASON_KEY_COMPROMISE ) fputs( "key_compromise ", fp ); 
    if( reason & KSBA_CRLREASON_CA_COMPROMISE ) fputs( "ca_compromise ", fp ); 
    if( reason & KSBA_CRLREASON_AFFILIATION_CHANGED ) fputs( "affiliation_changed ", fp ); 
    if( reason & KSBA_CRLREASON_SUPERSEDED ) fputs( "superseeded", fp ); 
    if( reason & KSBA_CRLREASON_CESSATION_OF_OPERATION ) fputs( "cessation_of_operation", fp ); 
    if( reason & KSBA_CRLREASON_CERTIFICATE_HOLD ) fputs( "certificate_hold", fp ); 
    if( reason & KSBA_CRLREASON_REMOVE_FROM_CRL ) fputs( "remove_from_crl", fp ); 
    fputs(") from ", fp );

    print_time( t, fp ); fputs( "\n", fp );
  } while( ( rc = cursorp->c_get( cursorp, &key, &val, DB_NEXT ) ) == 0 );

  return (rc==DB_NOTFOUND)?0:rc;
}

int 
crl_cache_list( CrlCache* c, FILE* fp ) 
{
  CrlCacheEntry* entry;
  int ret = 0;
  for( entry  = c->entries; entry; entry=entry->next ) {
    if( ( ret = list_one_crl_entry( entry, fp ) ) ) return ret;
  }
  return ret;
}

int
crl_cache_load( CrlCache* c, const char* filename, ASSUAN_CONTEXT ctx )
{
  struct stat sbuf;
  size_t size, tmp_size;
  char* buf;
  FILE* fp;
  int rc;

  if( stat( filename, &sbuf ) ) {
    log_error( "stat( %s ) failed, error is %s\n", filename, strerror( errno ) );
    return CRL_GENERAL_ERROR;
  }
  
  size = sbuf.st_size;
  buf = xmalloc( sizeof( char )*size );
  
  fp = fopen( filename, "r" );
  if( !fp ) {
    log_error( "fopen( %s ) failed, error is %s\n", filename, strerror( errno ) );
    xfree(buf);
    return CRL_GENERAL_ERROR;
  }

  if( size != (tmp_size = fread( buf, sizeof(char), size, fp )) ) {
    log_error( "fread return %d. Size of file is %d\n", tmp_size, size );
    xfree(buf);
    return CRL_GENERAL_ERROR;
  }

  rc = crl_cache_insert( c, filename, buf, size, ctx );
  xfree(buf);
  return rc;
}

